"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.qTreeVec = void 0;
const vector_1 = require("@4bitlabs/vector");
const bounds_1 = require("./bounds");
class qTreeVec {
    constructor([x0, y0, x1, y1], options, boundsFn, quadCache, depth = 0) {
        const { maxDepth = 7, maxChildren = 10 } = options;
        this.maxDepth = maxDepth;
        this.maxChildren = maxChildren;
        this.boundsFn = boundsFn;
        this.quadCache = quadCache;
        this.depth = depth;
        const midX = (x0 + x1) / 2;
        const midY = (y0 + y1) / 2;
        this.boundsVec = new vector_1.Vector(Float64Array, {
            initialLength: 16,
            initialCapacity: 64,
        });
        this.boundsVec.setN(0, [x0, y0, midX, midY]);
        this.boundsVec.setN(4, [midX, y0, x1, midY]);
        this.boundsVec.setN(8, [x0, midY, midX, y1]);
        this.boundsVec.setN(12, [midX, midY, x1, y1]);
        this.bounds = new WeakMap();
        this.childAreas = [
            this.boundsVec.subarray(0, 4),
            this.boundsVec.subarray(4, 8),
            this.boundsVec.subarray(8, 12),
            this.boundsVec.subarray(12, 16),
        ];
        this.isSplit = false;
        this.children = [null, null, null, null];
        this.items = new Set();
    }
    clear() {
        this.items.clear();
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            child.clear();
            this.children[i] = null;
        }
    }
    insert(item) {
        const bounds = this.boundsFn(item);
        this._insert(item, bounds);
    }
    /** @internal */
    _insert(item, bounds) {
        if (!this.isSplit && this.items.size + 1 > this.maxChildren) {
            this.isSplit = true;
            for (const i of this.items)
                if (this.tryChildInsert(i, bounds))
                    this.items.delete(i);
        }
        if (this.isSplit && this.depth + 1 < this.maxDepth) {
            if (this.tryChildInsert(item, bounds))
                return;
        }
        const idx = this.boundsVec.length;
        this.boundsVec.pushN(bounds);
        this.bounds.set(item, this.boundsVec.subarray(idx, idx + 4));
        this.items.add(item);
        this.quadCache.set(item, this);
    }
    tryChildInsert(item, bounds) {
        for (let i = 0; i < 4; i++) {
            const childRect = this.childAreas[i];
            if ((0, bounds_1.contains)(childRect, bounds)) {
                let child = this.children[i];
                if (!child) {
                    this.children[i] = child = new qTreeVec(childRect, { maxDepth: this.maxDepth }, this.boundsFn, this.quadCache, this.depth + 1);
                }
                child._insert(item, bounds);
                return true;
            }
        }
        return false;
    }
    remove(item) {
        if (this.items.has(item)) {
            this.items.delete(item);
            this.quadCache.delete(item);
            // TODO handle cleanup
            return true;
        }
        const quad = this.quadCache.get(item);
        if (!quad)
            return false;
        return quad.remove(item);
    }
    collectAll(result = []) {
        for (const r of this.items)
            result.push(r);
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            child.collectAll(result);
        }
        return result;
    }
    collect(area, result = []) {
        var _a;
        for (const item of this.items) {
            const rect = (_a = this.bounds.get(item)) !== null && _a !== void 0 ? _a : error('bounds');
            if ((0, bounds_1.overlaps)(area, rect))
                result.push(item);
        }
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            const childArea = this.childAreas[i];
            if ((0, bounds_1.contains)(area, childArea))
                child.collectAll(result);
            else if ((0, bounds_1.overlaps)(area, childArea))
                child.collect(area, result);
        }
        return result;
    }
    *all() {
        yield* this.items;
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            yield* child.all();
        }
    }
    *search(area) {
        var _a;
        for (const item of this.items) {
            const rect = (_a = this.bounds.get(item)) !== null && _a !== void 0 ? _a : error('bounds');
            if ((0, bounds_1.overlaps)(area, rect))
                yield item;
        }
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            const childArea = this.childAreas[i];
            if ((0, bounds_1.contains)(area, childArea))
                yield* child.all();
            else if ((0, bounds_1.overlaps)(area, childArea))
                yield* child.search(area);
        }
    }
    size() {
        let count = this.items.size;
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            count += child.size();
        }
        return count;
    }
}
exports.qTreeVec = qTreeVec;
function error(message) {
    throw new Error(message);
}
//# sourceMappingURL=q-tree-vec.js.map